from data.bag import Bag
from data.instance import Instance
from data.miml_dataset import MIMLDataset

values = [2, 7, 5.09, 1]
attr = {"hola1": 0, "hola2": 0, "hola3": 0, "label": 1}
instance1 = Instance(values, attr)
instance2 = Instance(values,attr)
instance3 = Instance(values,attr)
# print(instance1.data)
# instance1.show_instance()
bag = Bag(instance1, "bag1")
bag.add_instance(instance2)
bag.add_instance(instance3)

dataset = MIMLDataset()
dataset.add_bag(bag)
dataset.show_dataset()



from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report, hamming_loss
from sklearn.multioutput import MultiOutputClassifier

from datasets.load_dataset import load_dataset
from transformation.mimlTOml.arithmetic import ArithmeticTransformation

dataset_train = load_dataset("../datasets/miml_birds_random_80train.arff", delimiter="'")
dataset_test = load_dataset("../datasets/miml_birds_random_20test.arff", delimiter="'")

arithmetic_transformation_train = ArithmeticTransformation(dataset_train)
arithmetic_transformation_test = ArithmeticTransformation(dataset_test)

X_train, y_train = arithmetic_transformation_train.transform_dataset()
X_test, y_test = arithmetic_transformation_test.transform_dataset()

classifier = MultiOutputClassifier(RandomForestClassifier(random_state=27))
classifier.fit(X_train, y_train)

# Predicciones
y_pred = classifier.predict(X_test)

# Evaluación del modelo
print("Reporte de clasificación:\n", classification_report(y_test, y_pred, zero_division=0))

# print("Y TEST:",y_test)
# print("Y Pred:",y_pred)

print('Hamming Loss: ', round(hamming_loss(y_test, y_pred), 2))



from abc import ABC, abstractmethod


class AbstractClassifier(ABC):

    @abstractmethod
    def train(self, training_data, training_labels):
        pass

    @abstractmethod
    def predict(self, test_data):
        pass

    @abstractmethod
    def evaluate(self, test_data, test_labels):
        pass



import numpy as np

from data.miml_dataset import MIMLDataset


class ArithmeticTransformation:
    """
    Class that performs an arithmetic transformation to convert a MIMLDataset class to numpy ndarrays.
    """

    def __init__(self, dataset: MIMLDataset):
        self.dataset = dataset

    def transform_dataset(self):
        """
        Transform the dataset to multilabel dataset converting each bag into a single instance being the value of each
        attribute the mean value of the instances in the bag.

        Returns
        -------

        X : {numpy ndarray} of shape (number of instances, number of attributes)
        Training vector

        Y : {numpy ndarray} of shape (number of instances, number of labels)
        Target vector relative to X.

        """
        x = np.empty(shape=(
            self.dataset.get_number_bags(), self.dataset.get_number_attributes() - self.dataset.get_number_labels()))
        y = np.empty(shape=(self.dataset.get_number_bags(), self.dataset.get_number_labels()))
        count = 0
        for keys, pattern in self.dataset.data.items():
            values = pattern.data[0:, 0:self.dataset.get_number_attributes() - self.dataset.get_number_labels()]
            labels = pattern.data[0:, self.dataset.get_number_attributes() - self.dataset.get_number_labels():]
            new_instance = np.mean(values, axis=0)
            x[count] = new_instance
            y[count] = labels[0]
            count += 1

        return x, y

    def transform_instance(self, key):
        """
        Transform the instances of a bag to a multilabel instance

        Parameters
        ----------
        key : string
            Key of the bag to be transformed to multilabel instance

        Returns
        -------
        instance : tuple
        Tuple of numpy ndarray with attribute values and labels

        """

    # TODO: Implementarlo





def arff_to_csv(file, delimiter="'"):
    """
    Convert MIML Arff to CSV.

    Parameters
    ----------
    file : string
        Filepath of the file to be converted

    delimiter : string ( ' or " ), default = '
        Delimiter used in arff file for the start and end of the bag values

    """

    arff = open(file)
    csv = open(file[:-5] + ".csv", "w")
    attrib = []
    flag = 0

    for line in arff:
        # Comprobamos que la cadena no contenga espacios en blanco a la izquierda ni que sea vacía
        line = line.lstrip()
        if line == "":
            continue

        if line.startswith("%") or line.startswith("@"):
            if line.startswith("@attribute"):
                if not (line.startswith("@attribute bag relational")):
                    attrib.append(line[line.find(" ") + 1:line.find(" ", line.find(" ") + 1)])

        else:
            if flag == 0:
                for x in range(len(attrib)):
                    if x == len(attrib) - 1:
                        csv.write(attrib[x] + "\n")
                    else:
                        csv.write(attrib[x] + ",")
                flag = 1

            # Eliminanos el salto de línea del final de la cadena
            line = line.strip("\n")

            # Asumimos que el primer elemento de cada instancia es el identificador de la bolsa
            key = line[0:line.find(",")]
            # print("Key: ", key)

            # Empiezan los datos de la bolsa cuando encontremos la primera '"' y terminan con la segunda '"'
            line = line[line.find(delimiter) + 1:]
            values = line[:line.find(delimiter, line.find(delimiter, line.find(delimiter)))]
            # Separamos los valores por instancias de la bolsa
            values = values.split("\\n")
            # print("Values ", values)

            # El resto de la cadena se trata de las etiquetas
            labels = line[line.find(delimiter, line.find(delimiter, line.find(delimiter))) + 2:]
            # print("Labels: ", labels)

            for v in values:
                csv.write(key + "," + v + "," + labels + "\n")
                # TODO: quizas separar en funciones para data y para atributos
                # TODO: intentar optimizar la funcion, aprovechar que la string del arff
                # es parecida, separar solo los values
                # TODO: control de errores

    arff.close()
    csv.close()



import numpy as np
from tabulate import tabulate

from data.instance import Instance


class Bag:

    def __init__(self, instance, key):
        # Crear numpy ndarray 3d con una instancia
        self.data = np.array(instance.data)
        self.key = key
        self.attributes = instance.attributes

    def get_number_instances(self):
        """
        Get numbers of instances of a bag

        Parameters
        ----------
        key : string
            Key of the bag

        Returns
        ----------
        numbers of instances: int
            Numbers of instances of a bag

        """
        return len(self.data)

    def get_instance(self, index):
        """

        Parameters
        ----------

        index : int
            Index of the instance in the bag

        Returns
        -------
        instance : tuple of ndarrays
            Tuple with attribute values and label of the instance

        """
        pass

    def get_attributes(self):
        return list(self.attributes.keys())

    def add_instance(self, instance):
        """

        Parameters
        ----------
        key : string
            Key of the bag
        values : ndarray
            Values of the instance to be inserted

        """
        # TODO: Check same length

        self.data = np.vstack((self.data, instance.data))

    def add_attribute(self, name, position, value=0):
        if position is None:
            position = len(self.data)
        pass

    def set_attribute(self, index, value):
        """
        Update value from attributes

        Parameters
        ----------
        key : string
            Bag key of the dataset

        attribute: string
            Attribute of the dataset

        value: float
            New value for the update
        """
        pass

    def set_attribute(self, name, value):
        pass

    def delete_instance(self, index):
        pass

    def delete_attribute(self, position):
        pass

    def delete_attribute(self, name):
        pass

    def show_bag(self):
        # TODO: Check

        table = [[self.key] + self.get_attributes()]
        count = 0
        for instance in self.data:
            table.append([count] + list(instance))
            count += 1
        # table = [['col 1', 'col 2', 'col 3', 'col 4'], [1, 2222, 30, 500], [4, 55, 6777, 1]]
        # print(tabulate(table, headers='firstrow', tablefmt='fancy_grid'))
        # print(tabulate([key], tablefmt="grid"))
        print(tabulate(table, headers='firstrow', tablefmt="grid", numalign="center"))
